# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from typing import List

from pydantic import BaseModel
from pydantic import validator
from vdk.api.job_input import IJobInput
from vdk.plugin.impala.templates.template_arguments_validator import (
    TemplateArgumentsValidator,
)


class LoadVersionedParams(BaseModel):
    target_schema: str
    target_table: str
    source_schema: str
    source_view: str
    id_column: str
    value_columns: List[str]
    tracked_columns: List[str]
    updated_at_column: str = "updated_at"
    sk_column: str = "sk"
    active_from_column: str = "active_from"
    active_to_column: str = "active_to"
    active_to_max_value: str = "9999-12-31"

    @validator("tracked_columns", allow_reuse=True)
    def passwords_match(cls, tracked_columns, values, **kwargs):
        value_columns = values.get("value_columns")
        if type(value_columns) == list and not tracked_columns:
            raise ValueError("The list must contain at least one column")
        if type(value_columns) == list == type(value_columns) and not set(
            tracked_columns
        ) <= set(value_columns):
            raise ValueError(
                "All elements in the list must be also present in `value_columns`"
            )
        return tracked_columns


class LoadVersioned(TemplateArgumentsValidator):
    TemplateParams = LoadVersionedParams

    def __init__(self) -> None:
        super().__init__()

    def _validate_args(self, args: dict) -> dict:
        args = super()._validate_args(args)
        return dict(
            **args,
            value_columns_str=", ".join(
                [f"`{column}`" for column in args["value_columns"]]
            ),
            hash_expr_str=",\n".join(
                [
                    f"            COALESCE(CAST(`{column}` AS STRING), '#')"
                    for column in args["tracked_columns"]
                ]
            ).lstrip(),
        )


def run(job_input: IJobInput):
    LoadVersioned().get_validated_args(job_input, job_input.get_arguments())



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import re
import time

from vdk.internal.builtin_plugins.connection.recovery_cursor import RecoveryCursor
from vdk.internal.core import errors


class ImpalaErrorHandler:
    def __init__(self, log=None, num_retries=5):
        """
        This module offers error handling and error recovery for an Impala connection.

        :param log: a logger object, creates a new one if one isn't passed as a parameter
        :param num_retries: number of times a query will be re-tried
        """
        if not log:
            log = logging.getLogger(__name__)
        self._log = log

        self._num_retries = num_retries

    def handle_error(
        self, caught_exception: Exception, recovery_cursor: RecoveryCursor
    ) -> bool:
        """
        Handles an exception/error thrown by Impala.
        It examines the error and depending on which case it matches it will
        take measures to fix the problem and/or re-try the query.

        For example, let's consider a "File Not Found" error.
        Queries can fail when they are querying for tables which have been recently compacted.
        In such a case, the following steps are undertaken:
            1. Examine exception to detect if this is the case;
            2. If it is, then issue a refresh for the table whose files could not be found;
            3. Re-try the query.

        :param caught_exception: the caught Impala exception
        :param recovery_cursor: cursor that can be used to execute recovery queries and query retries against Impala
        :return: true if and only if the exception was handled successfully and the query passed has succeeded.
                Otherwise return false - the caught_exception was not handled.
                This method will raise an exception in the following scenarios and assuming handling of the error
                produced another exception:
                 1. Exceptions are of the same type but have different args.
                 2. Exceptions are different types.
                Moreover, the root cause stack trace will appear in the error message if an exception is raised.
                Otherwise stacktrace gets changed and becomes confusing.
        """
        if self._handle_should_not_retry_error(caught_exception):
            return False

        if self._is_pool_error(caught_exception):
            errors.log_and_throw(
                to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                log=self._log,
                what_happened="An Impala Pool Error occured: " + str(caught_exception),
                why_it_happened="Review the contents of the exception.",
                consequences="The queries will not be executed.",
                countermeasures=(
                    "Optimise the executed queries. Alternatively, make sure that "
                    "the data job is not running too many queries in parallel."
                ),
            )

        is_handled = False
        # try to handle multiple failed to open file errors in one query for different tables
        current_exception = caught_exception
        while recovery_cursor.get_retries() < self._num_retries:
            self._log.info(
                f"Try ({(recovery_cursor.get_retries() + 1)} of {self._num_retries}) "
                f"to handle exception {current_exception}"
            )
            try:
                is_handled = self._handle_exception(current_exception, recovery_cursor)
                break
            except Exception as new_exception:
                self._log.warning(
                    "Failed to handle exception due to: "
                    + errors.get_exception_message(new_exception)
                )
                current_exception = new_exception
        # Throw the latest exception only if different from the initial exception.
        if not is_handled and (
            type(current_exception) is not type(caught_exception)
            or current_exception.args != caught_exception.args
        ):
            self._log.warning(
                "Query failed with an exception. We tried handling and resolving the original error: "
                + f"({(type(caught_exception))}) by retrying the query "
                f"({(recovery_cursor.get_retries())}) times. "
                + "Another exception occurred during handling of the above mentioned error. "
                + "Throwing the exception which occurred during the latest retry of the original error. "
                + "For more information on both errors check the stack trace bellow: "
            )
            raise current_exception
        return is_handled

    def _handle_exception(self, exception, recovery_cursor) -> bool:
        # The call to the _handle_impala_network_error() method below needs to stay at the end, as
        # it catches all OperationalError cases not caught in the preceding method calls.
        return (
            self._handle_hdfs_failed_to_open_file_error(exception, recovery_cursor)
            or self._handle_metadata_exception_with_backoff(exception, recovery_cursor)
            or self._handle_metadata_exception_with_invalidate_and_backoff(
                exception, recovery_cursor
            )
            or self._handle_metadata_exception_with_invalidate_and_retry(
                exception, recovery_cursor
            )
            or self._handle_impala_network_error(exception, recovery_cursor)
        )

    def _handle_should_not_retry_error(self, exception) -> bool:
        if errors.exception_matches(
            exception,
            classname_with_package="impala.error.OperationalError",
            exception_message_matcher_regex=".*expired due to execution time limit.*",
        ):
            self._log.info(
                "Execution of query exceeded impala time limit. "
                "This is most likely because the query is resource heavy and needs optimisation. "
                "The query would not be retried, in order to avoid increasing the load on the database."
            )
            return True

        return False

    def _handle_hdfs_failed_to_open_file_error(
        self, exception, recovery_cursor
    ) -> bool:
        if errors.exception_matches(
            exception, "impala.error.OperationalError", ".*Failed to open HDFS.*"
        ) or errors.exception_matches(
            exception, "impala.error.HiveServer2Error", ".*Failed to open HDFS.*"
        ):
            regex = ".*/user/hive/warehouse/([^/]*).db/([^/]*)"
            matcher = re.compile(pattern=regex, flags=re.DOTALL)
            results = matcher.search(errors.get_exception_message(exception))
            self._log.info(
                "Detected query failing with Failed to find file error. WIll try to autorecover."
            )
            if results and len(results.groups()) == 2:
                database = results.group(1)
                table = results.group(2)
                self._log.info(
                    "Query failed with with Failed to find file error. "
                    "This is most likely due to delay metadata sync from compaction. "
                    f"We are issuing refresh statement for the offending table: {database}.{table}"
                    " and retrying the query"
                    f"Error was: {exception.__class__}: {str(exception)}"
                )
                # Try refreshing the table metadata several times,
                # and if the issue is not fixed invalidate the metadata.
                if recovery_cursor.get_retries() > 3:
                    recovery_cursor.execute(
                        "invalidate metadata `" + database + "`.`" + table + "`"
                    )
                else:
                    recovery_cursor.execute(
                        "refresh `" + database + "`.`" + table + "`"
                    )
                    time.sleep(
                        2 ** recovery_cursor.get_retries() * 30
                    )  # exponential backoff 30s, 60s, 2m, 4m, 8m
                recovery_cursor.retry_operation()
                return True
            else:
                self._log.info(
                    f"Cannot auto-recover as it cannot detect table name in error: {errors.get_exception_message(exception)}"
                )
        return False

    def _handle_impala_network_error(self, exception, recovery_cursor) -> bool:
        if (
            errors.exception_matches(
                exception,
                classname_with_package="impala.error.OperationalError",
                exception_message_matcher_regex=".*.*",
            )
            or errors.exception_matches(
                exception,
                classname_with_package="impala.error.HiveServer2Error",
                exception_message_matcher_regex=".*Connection refused.*",
            )
            or errors.exception_matches(
                exception,
                classname_with_package="impala.error.HiveServer2Error",
                exception_message_matcher_regex=".*Failed after retrying 3 times.*",
            )
        ):
            self._log.info(
                "Query failed with network error. This is most likely a recoverable error. "
                "We are retrying the query. "
                f"Error was: {exception.__class__}: {str(exception)}"
            )
            # wait a little before retrying the query, giving time for the network to recover
            if recovery_cursor.get_retries() > 0:
                time.sleep(
                    2 ** recovery_cursor.get_retries() * 30
                )  # exponential backoff 60s, 2m, 4m, 8m
            # retry the query
            recovery_cursor.retry_operation()
            return True
        else:
            return False

    def _handle_metadata_exception_with_backoff(
        self, exception, recovery_cursor
    ) -> bool:
        if (
            errors.exception_matches(
                exception,
                classname_with_package="impala.error.HiveServer2Error",
                exception_message_matcher_regex=".*AnalysisException: Could not resolve column/field reference.*",
            )
            or errors.exception_matches(
                exception,
                classname_with_package="impala.error.HiveServer2Error",
                exception_message_matcher_regex=".*OutOfMemoryError: null.*",
            )
            or errors.exception_matches(
                exception,
                classname_with_package="impala.error.OperationalError",
                exception_message_matcher_regex=".*ImpalaRuntimeException: Error making 'alter_table' RPC to Hive.*",
            )
            or errors.exception_matches(
                exception,
                classname_with_package="impala.error.OperationalError",
                exception_message_matcher_regex=".*ImpalaRuntimeException: Error making 'updateTableColumnStatistics'.*",
            )
        ):

            sleep_seconds = 2 ** recovery_cursor.get_retries() * 30
            self._log.info(
                f"Query failed with: {exception.__class__} : {str(exception)}"
                f"Will sleep for {sleep_seconds} seconds and try the query again."
            )
            # wait a little before retrying the query, giving time for the metadata to sync
            time.sleep(sleep_seconds)  # exponential backoff 30s, 60s, 2m, 4m, 8m
            # retry the query
            recovery_cursor.retry_operation()
            return True
        else:
            return False

    def _handle_metadata_exception_with_invalidate_and_backoff(
        self, exception, recovery_cursor
    ) -> bool:
        pattern_for_the_table_name = None

        if errors.exception_matches(
            exception,
            classname_with_package="impala.error.OperationalError",
            exception_message_matcher_regex=".*TableLoadingException: Error loading metadata for table.*",
        ):
            pattern_for_the_table_name = (
                ".*TableLoadingException: Error loading metadata for table: (\\S+).*"
            )

        if errors.exception_matches(
            exception,
            classname_with_package="impala.error.HiveServer2Error",
            exception_message_matcher_regex=".*TableLoadingException: Error loading metadata for table.*",
        ):
            pattern_for_the_table_name = (
                ".*TableLoadingException: Error loading metadata for table: (\\S+).*"
            )

        if errors.exception_matches(
            exception,
            classname_with_package="impala.error.HiveServer2Error",
            exception_message_matcher_regex=".*LocalCatalogException: Could not load table.*",
        ):
            pattern_for_the_table_name = (
                ".*LocalCatalogException: Could not load table (\\S+) from metastore.*"
            )

        if errors.exception_matches(
            exception,
            classname_with_package="impala.error.HiveServer2Error",
            exception_message_matcher_regex=".*AnalysisException: Could not resolve table reference.*",
        ):
            pattern_for_the_table_name = (
                ".*AnalysisException: Could not resolve table reference: '(\\S+)'.*"
            )

        if (
            pattern_for_the_table_name
        ):  # then one of the exceptions matched and we can handle with invalidate
            sleep_seconds = 2 ** recovery_cursor.get_retries() * 30
            self._log.info(
                f"Query failed with: {exception.__class__} : {str(exception)}"
                f"Will sleep for {sleep_seconds} seconds, will issue invalidate metadata on the table and try the query again."
            )

            # Get the fully qualified table name from the exception message.
            matcher = re.compile(pattern=pattern_for_the_table_name)
            results = matcher.search(errors.get_exception_message(exception))
            if results and len(results.groups()) == 1:
                fully_qualified_table_name = results.group(1)
                # invalidate the metadata for the missing table
                recovery_cursor.execute(
                    f"invalidate metadata {fully_qualified_table_name}"
                )
                # wait a little before retrying the query, relaxing the stress on the metastore service
                self._log.info(
                    f"Sleeping for {sleep_seconds} seconds before retrying the query ..."
                )
                time.sleep(sleep_seconds)  # exponential backoff 30s, 60s, 2m, 4m, 8m
                # retry the query
                recovery_cursor.retry_operation()
                return True
        return False

    # The method is meant to handle only failed queries, whose metadata needs to be invalidated, and that should
    # be retried only once to avoid overloading the database.
    def _handle_metadata_exception_with_invalidate_and_retry(
        self, exception, recovery_cursor
    ) -> bool:
        def check_exception_and_get_pattern_for_table_name(exception_to_match):
            pattern = None
            if errors.exception_matches(
                exception_to_match,
                classname_with_package="impala.error.HiveServer2Error",
                exception_message_matcher_regex=".*AnalysisException: Table does not exist:.*",
            ):
                pattern = ".*AnalysisException: Table does not exist: (\\S+)"
            if errors.exception_matches(
                exception_to_match,
                classname_with_package="impala.error.HiveServer2Error",
                exception_message_matcher_regex=".*AlreadyExistsException: Table.*",
            ):
                pattern = ".*AlreadyExistsException: Table (\\S+).*"
            if errors.exception_matches(
                exception_to_match,
                classname_with_package="impala.error.HiveServer2Error",
                exception_message_matcher_regex=".*LocalCatalogException: Could not load.*",
            ):
                pattern = ".*LocalCatalogException: Could not load (\\S+).*"
            return pattern

        def get_fully_qualified_table_name(exception_to_match: Exception):
            pattern_for_the_table_name = check_exception_and_get_pattern_for_table_name(
                exception_to_match
            )
            if pattern_for_the_table_name:
                # Get the fully qualified table name from the exception message.
                matcher = re.compile(pattern=pattern_for_the_table_name)
                results = matcher.search(
                    errors.get_exception_message(exception_to_match)
                )
                if results and len(results.groups()) == 1:
                    fully_qualified_table_name = results.group(1)

                    # Data Mart name may not be available from the exception message, so we need to extract it
                    # from the query.
                    if "." not in fully_qualified_table_name:
                        try:
                            schema_match = re.search(
                                r"create (?:table|view)\b\s`?(\w+)`?\.{}".format(
                                    fully_qualified_table_name
                                ),
                                recovery_cursor.get_managed_operation().get_operation(),
                                re.IGNORECASE,
                            ).group(1)
                            fully_qualified_table_name = (
                                f"{schema_match}.{fully_qualified_table_name}"
                            )
                        except AttributeError:
                            self._log.info(
                                "Could not retrieve table name %s from provided query."
                                % fully_qualified_table_name
                            )

                            return None

                    return fully_qualified_table_name
                return None
            return None

        detected_table = get_fully_qualified_table_name(exception)
        if detected_table:
            # then one of the exceptions matched and we can handle with invalidate
            sleep_seconds = 10
            self._log.info(
                f"Query failed with: {exception.__class__} : {str(exception)}"
                f"Will sleep for {sleep_seconds} seconds, will issue invalidate"
                " metadata on the table and try the query again."
            )

            # invalidate the metadata for the table. This is necessary in case the metadata for the table
            # has not been propagated before the query is executed again.
            recovery_cursor.execute(f"invalidate metadata {detected_table}")
            # wait a little before retrying the query, relaxing the stress on the metastore service
            self._log.info(
                f"Sleeping for {sleep_seconds} seconds before retrying the query ..."
            )
            time.sleep(sleep_seconds)  # sleep for 20s
            # retry the query
            try:
                recovery_cursor.retry_operation()
            except Exception as new_exception:
                new_detected_table = get_fully_qualified_table_name(new_exception)
                if new_detected_table == detected_table:
                    self._log.info(
                        f"We tried to invalidate metadata for {detected_table} "
                        f"and retry query but it did not seem to work."
                        "We give up and will not try again"
                    )
                    return False
                # new exception is different so we re-raise so it can be handled again.
                raise

            return True
        return False

    @staticmethod
    def _is_pool_error(exception: Exception) -> bool:
        """
        Check for impala pool errors.
        """
        return errors.exception_matches(
            exception,
            classname_with_package=".*OperationalError.*",
            exception_message_matcher_regex=".*queue full.*",
        ) or errors.exception_matches(
            exception,
            classname_with_package=".*OperationalError.*",
            exception_message_matcher_regex=".*admission for timeout ms in pool.*",
        )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.internal.core import errors


def is_impala_user_error(received_exception):
    if errors.exception_matches(
        received_exception,
        classname_with_package="impala.error.HiveServer2Error",
        exception_message_matcher_regex="^AnalysisException: This Impala daemon is not ready to accept user requests.*",
    ):
        return False

    return (
        _is_syntax_error(received_exception)
        or _is_udf_error(received_exception)
        or _is_authorization_error(received_exception)
        or _is_query_limit_exceeded(received_exception)
        or _is_impala_memory_limit_exceeded(received_exception)
        or _is_quota_exceeded(received_exception)
        or _is_other_query_error(received_exception)
        or _is_property_unsupported_value_error(received_exception)
    )


def _is_syntax_error(exception):
    return (
        errors.exception_matches(
            exception,
            classname_with_package="impala.error.HiveServer2Error",
            exception_message_matcher_regex="^AnalysisException.*",
        )
        or errors.exception_matches(
            exception,
            classname_with_package="impala.error.HiveServer2Error",
            exception_message_matcher_regex="^ParseException: Syntax error.*",
        )
        or errors.exception_matches(
            exception,
            classname_with_package="impala.error.HiveServer2Error",
            exception_message_matcher_regex="^Could not compile regexp pattern.*",
        )
        or errors.exception_matches(
            exception,
            classname_with_package="impala.error.HiveServer2Error",
            exception_message_matcher_regex="^TableNotFoundException.*",
        )
        or errors.exception_matches(
            exception,
            classname_with_package="impala.error.HiveServer2Error",
            exception_message_matcher_regex=".*ArrayIndexOutOfBoundsException:.*",
        )
    )


def _is_udf_error(exception):
    return errors.exception_matches(
        exception,
        classname_with_package="impala.error.OperationalError",
        exception_message_matcher_regex="^UDF ERROR:.*",
    )


def _is_authorization_error(exception):
    return errors.exception_matches(
        exception,
        classname_with_package="impala.error.HiveServer2Error",
        exception_message_matcher_regex="^AuthorizationException:.*",
    )


def _is_query_limit_exceeded(exception):
    return (
        # Error specified below occurs when query exceeds 1h time limit
        errors.exception_matches(
            exception,
            classname_with_package="impala.error.OperationalError",
            exception_message_matcher_regex="^Cancelled from Impala's debug web interface$",
        )
        or errors.exception_matches(
            exception,
            classname_with_package="impala.error.OperationalError",
            exception_message_matcher_regex="^.*expired due to execution time limit.*$",
        )
    )


def _is_other_query_error(exception):
    return errors.exception_matches(
        exception,
        classname_with_package="impala.error.OperationalError",
        exception_message_matcher_regex=".*Cannot perform hash join at node with id.*",
    )


def _is_property_unsupported_value_error(exception):
    return errors.exception_matches(
        exception,
        classname_with_package="ValueError",
        exception_message_matcher_regex=".*[pP]roperty.*is of unsupported type.*",
    )


def _is_impala_memory_limit_exceeded(exception):
    return errors.exception_matches(
        exception,
        classname_with_package="impala.error.OperationalError",
        exception_message_matcher_regex=".*Scratch space limit.*",
    ) or errors.exception_matches(
        exception,
        classname_with_package="impala.error.OperationalError",
        exception_message_matcher_regex=".*Memory limit exceeded.*",
    )


def _is_quota_exceeded(exception):
    return errors.exception_matches(
        exception,
        classname_with_package="impala.error.OperationalError",
        exception_message_matcher_regex=".*DiskSpace quota of .* is exceeded.*",
    )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
import os
import pathlib
from typing import List

import click
from tabulate import tabulate
from vdk.api.plugin.hook_markers import hookimpl
from vdk.api.plugin.plugin_registry import HookCallResult
from vdk.api.plugin.plugin_registry import IPluginRegistry
from vdk.internal.builtin_plugins.connection.decoration_cursor import DecorationCursor
from vdk.internal.builtin_plugins.connection.recovery_cursor import RecoveryCursor
from vdk.internal.builtin_plugins.run.job_context import JobContext
from vdk.internal.builtin_plugins.run.step import Step
from vdk.internal.core.config import ConfigurationBuilder
from vdk.internal.core.errors import ErrorMessage
from vdk.internal.core.errors import UserCodeError
from vdk.plugin.impala.impala_configuration import add_definitions
from vdk.plugin.impala.impala_configuration import ImpalaPluginConfiguration
from vdk.plugin.impala.impala_connection import ImpalaConnection
from vdk.plugin.impala.impala_error_classifier import is_impala_user_error
from vdk.plugin.impala.impala_error_handler import ImpalaErrorHandler


def _connection_by_configuration(configuration: ImpalaPluginConfiguration):
    return ImpalaConnection(
        host=configuration.host(),
        port=configuration.port(),
        database=configuration.database(),
        timeout=configuration.timeout(),
        use_ssl=configuration.use_ssl(),
        ca_cert=configuration.ca_cert(),
        auth_mechanism=configuration.auth_mechanism(),
        user=configuration.user(),
        password=configuration.password(),
        kerberos_service_name=configuration.kerberos_service_name(),
        krb_host=configuration.krb_host(),
        use_http_transport=configuration.use_http_transport(),
        http_path=configuration.http_path(),
        auth_cookie_names=configuration.auth_cookie_names(),
        retries=configuration.retries(),
    )


@click.command(name="impala-query", help="executes SQL query against Impala")
@click.option("-q", "--query", type=click.STRING, required=True)
@click.pass_context
def impala_query(ctx: click.Context, query):
    impala_cfg = ImpalaPluginConfiguration(ctx.obj.configuration)
    click.echo(tabulate(_connection_by_configuration(impala_cfg).execute_query(query)))


class ImpalaPlugin:
    @staticmethod
    @hookimpl
    def vdk_configure(config_builder: ConfigurationBuilder) -> None:
        add_definitions(config_builder)

    @staticmethod
    @hookimpl
    def vdk_command_line(root_command: click.Group):
        root_command.add_command(impala_query)

    @hookimpl
    def initialize_job(self, context: JobContext) -> None:
        self._impala_cfg = ImpalaPluginConfiguration(
            context.core_context.configuration
        )  # necessary for the query decoration hook
        context.connections.add_open_connection_factory_method(
            "IMPALA",
            lambda: _connection_by_configuration(self._impala_cfg),
        )

        context.templates.add_template(
            "load/dimension/scd1", pathlib.Path(get_job_path("load/dimension/scd1"))
        )

        context.templates.add_template(
            "scd1", pathlib.Path(get_job_path("load/dimension/scd1"))
        )

        context.templates.add_template(
            "load/dimension/scd2", pathlib.Path(get_job_path("load/dimension/scd2"))
        )

        context.templates.add_template(
            "scd2_simple", pathlib.Path(get_job_path("load/dimension/scd2"))
        )

        context.templates.add_template(
            "load/fact/snapshot", pathlib.Path(get_job_path("load/fact/snapshot"))
        )

        context.templates.add_template(
            "periodic_snapshot", pathlib.Path(get_job_path("load/fact/snapshot"))
        )

        context.templates.add_template(
            "load/versioned", pathlib.Path(get_job_path("load/versioned"))
        )
        context.templates.add_template(
            "scd2", pathlib.Path(get_job_path("load/versioned"))
        )

    @staticmethod
    @hookimpl(hookwrapper=True, tryfirst=True)
    def run_step(context: JobContext, step: Step) -> None:
        out: HookCallResult
        out = yield

        if out.result.exception:
            if is_impala_user_error(out.result.exception):
                raise UserCodeError(
                    ErrorMessage(
                        summary="Error occurred.",
                        what=f"Error occurred. Exception message: {out.result.exception}",
                        why="Review exception for details.",
                        consequences="Data Job execution will not continue.",
                        countermeasures="Review exception for details.",
                    )
                ) from out.result.exception

    @staticmethod
    @hookimpl
    def db_connection_recover_operation(recovery_cursor: RecoveryCursor) -> None:
        impala_error_handler = ImpalaErrorHandler()

        if impala_error_handler.handle_error(
            recovery_cursor.get_exception(), recovery_cursor
        ):
            logging.getLogger(__name__).info(
                "Error handled successfully! Query execution has succeeded."
            )
        else:
            raise recovery_cursor.get_exception()

    @hookimpl(tryfirst=True)
    def db_connection_decorate_operation(self, decoration_cursor: DecorationCursor):
        if self._impala_cfg.sync_ddl():
            decoration_cursor.execute("SET SYNC_DDL=True")
        if self._impala_cfg.query_pool():
            decoration_cursor.execute(
                f"SET REQUEST_POOL='{self._impala_cfg.query_pool()}'"
            )


@hookimpl
def vdk_start(plugin_registry: IPluginRegistry, command_line_args: List):
    plugin_registry.load_plugin_with_hooks_impl(ImpalaPlugin(), "impala-plugin")


def get_jobs_parent_directory() -> pathlib.Path:
    current_dir = pathlib.Path(os.path.dirname(os.path.abspath(__file__)))
    jobs_dir = current_dir.joinpath("templates")
    return jobs_dir


def get_job_path(job_name: str) -> str:
    """Get the path of the test data job returned as string so it can be passed easier as cmd line args"""
    return str(get_jobs_parent_directory().joinpath(job_name))



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from pydantic import BaseModel
from vdk.api.job_input import IJobInput
from vdk.plugin.impala.templates.template_arguments_validator import (
    TemplateArgumentsValidator,
)


class SlowlyChangingDimensionType2Params(BaseModel):
    target_schema: str
    target_table: str
    source_schema: str
    source_view: str
    start_time_column: str
    end_time_column: str
    end_time_default_value: str
    surrogate_key_column: str
    id_column: str


class SlowlyChangingDimensionType2(TemplateArgumentsValidator):
    TemplateParams = SlowlyChangingDimensionType2Params

    def __init__(self) -> None:
        super().__init__()


def run(job_input: IJobInput):
    SlowlyChangingDimensionType2().get_validated_args(
        job_input, job_input.get_arguments()
    )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging
from collections import OrderedDict

import pyarrow
from vdk.internal.core import errors
from vdk.plugin.impala import impala_error_classifier
from vdk.plugin.impala.impala_connection import ImpalaConnection


class ImpalaHelper:
    def __init__(self, db_connection: ImpalaConnection) -> None:
        self._log = logging.getLogger(__name__)
        self._db_connection = db_connection

    def get_table_description(self, table_name):
        self._log.debug(f"Retrieving details for table {table_name}.")
        try:
            return self._db_connection.execute_query(f"DESCRIBE formatted {table_name}")
        except Exception as e:
            if impala_error_classifier._is_authorization_error(e):
                errors.log_and_throw(
                    to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                    log=self._log,
                    what_happened=f"Data loading into table {table_name} has failed.",
                    why_it_happened=(
                        f"You are trying to load data into a table which you do not have access to or it does not "
                        f"exist: {table_name}."
                    ),
                    consequences="Data load will be aborted.",
                    countermeasures="Make sure that the destination table exists and you have access to it.",
                )
            else:
                raise e

    def __get_table_schema(self, table_description, section_start="#", second_end="#"):
        """
        Gets column names and data types from Impala table.
        It would be a lot more easier to execute pure describe statement, but then we will execute 2 describe statements
        to get the table schema and check if the table is stored as parquet
        Will return the full table schema including partition columns order the same way as in Impala
        """
        self._log.debug("Retrieving destination table schema.")
        column_name_to_column_type_map = OrderedDict()
        is_in_columns_section = False
        searched_section_ended = False
        searched_sectioned_started = False
        for (
            column_name,
            column_type,
            _,
        ) in table_description:  # 3rd value is column comment
            if column_name is None or column_name.strip() == "":
                continue
            if column_name.startswith(section_start):  # new section begins
                searched_sectioned_started = True
            if searched_sectioned_started and not is_in_columns_section:
                if column_name.strip() == "# col_name":  # column info follows
                    is_in_columns_section = True
                else:
                    is_in_columns_section = False
                continue
            if searched_sectioned_started:
                if column_name.startswith(second_end):
                    searched_section_ended = True
            if is_in_columns_section and not searched_section_ended:
                column_name_to_column_type_map[
                    column_name.strip()
                ] = column_type.strip()
        return column_name_to_column_type_map

    def get_table_columns(self, table_description):
        """
        :param table_description: result of #get_table_description
        :return: dict with column name and type
        """
        return self.__get_table_schema(table_description, section_start="#")

    def get_table_partitions(self, table_description):
        """
        :param table_description: result of #get_table_description
        :return: dict with partition name and type
        """
        return self.__get_table_schema(
            table_description, section_start="# Partition Information"
        )

    def ensure_table_format_is_parquet(self, table_name, table_description):
        for key, value, _ in table_description:  # 3rd value is column comment
            # Input format of parquet table is "org.apache.hadoop.hive.ql.io.parquet.MapredParquetInputFormat"
            if key is not None and key.strip() == "InputFormat:":
                if "parquet" in value:  # table is stored as parquet
                    return
                else:  # table is not stored as parquet
                    errors.log_and_throw(
                        to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                        log=self._log,
                        what_happened="Data loading has failed.",  # FIXME: this is too specific
                        why_it_happened=(
                            f"You are trying to load data into a table {table_name} with an unsupported format. "
                            f"Currently only Parquet table format is supported."
                        ),
                        consequences="Data load will be aborted.",  # FIXME: this is too specific
                        countermeasures=(
                            "Make sure that the destination table is stored as parquet: "
                            "https://www.cloudera.com/documentation/enterprise/5-11-x/topics/impala_parquet.html"
                            "#parquet_ddl"
                        ),
                    )
        # TODO once there is more robust loading implemented the below error can be removed. We can try to load even if
        # we cannot determine the table storage type
        errors.log_and_throw(
            to_be_fixed_by=errors.ResolvableBy.PLATFORM_ERROR,
            log=self._log,
            what_happened="Cannot determine the target table file format, which is needed to load data into it.",
            why_it_happened="There's a bug in VDK code.",
            consequences="Application will exit.",
            countermeasures="Report this bug to versatile data kit team.",
        )

    def generate_parquet_schema_from_table_schema(self, table_columns):
        """
        Builds the parquet schema based on the column types and order in the target table, in order to ensure the new file
        will be compatible with the table
        """
        self._log.debug("Generating parquet file schema from table schema.")
        impala_type_to_pyarrow_type_map = {
            "string": pyarrow.string(),
            "boolean": pyarrow.bool_(),
            "double": pyarrow.float64(),
            "float": pyarrow.float32(),
            "int": pyarrow.int32(),
            "bigint": pyarrow.int64(),
            "timestamp": pyarrow.timestamp("ns"),
        }
        # including the decimal types in the map
        for precision_value in range(1, 39):
            for scale_value in range(0, precision_value + 1):
                impala_type_to_pyarrow_type_map[
                    f"decimal({precision_value},{scale_value})"
                ] = pyarrow.decimal128(precision_value, scale_value)

        parquet_schema = []
        for column_name, column_type in table_columns.items():
            parquet_schema.append(
                (column_name, impala_type_to_pyarrow_type_map[column_type])
            )
        return pyarrow.schema(parquet_schema)

    def get_parquet_schema(self, table):
        table_description = self.get_table_description(table)
        self.ensure_table_format_is_parquet(table, table_description)
        table_columns = self.get_table_columns(table_description)
        return self.generate_parquet_schema_from_table_schema(table_columns)

    @staticmethod
    def get_insert_sql_partition_clause(partitions):
        # https://docs.cloudera.com/documentation/enterprise/6/6.3/topics/impala_insert.html

        # NOTE: https://github.com/kayak/pypika looks interesting if we start having more complex query buildings
        sql = "PARTITION (" + ",".join("`" + p + "`" for p in partitions.keys()) + ")"
        return sql



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from vdk.internal.core.config import ConfigurationBuilder

IMPALA_HOST = "IMPALA_HOST"
IMPALA_PORT = "IMPALA_PORT"
IMPALA_DATABASE = "IMPALA_DATABASE"
IMPALA_TIMEOUT = "IMPALA_TIMEOUT"
IMPALA_USE_SSL = "IMPALA_USE_SSL"
IMPALA_CA_CERT = "IMPALA_CA_CERT"
IMPALA_AUTH_MECHANISM = "IMPALA_AUTH_MECHANISM"
IMPALA_USER = "IMPALA_USER"
IMPALA_PW = "IMPALA_PASSWORD"
IMPALA_KERBEROS_SERVICE_NAME = "IMPALA_KERBEROS_SERVICE_NAME"
IMPALA_KRB_HOST = "IMPALA_KRB_HOST"
IMPALA_USE_HTTP_TRANSPORT = "IMPALA_USE_HTTP_TRANSPORT"
IMPALA_HTTP_PATH = "IMPALA_HTTP_PATH"
IMPALA_AUTH_COOKIE_NAMES = "IMPALA_AUTH_COOKIE_NAMES"
IMPALA_RETRIES = "IMPALA_RETRIES"
IMPALA_SYNC_DDL = "IMPALA_SYNC_DDL"
IMPALA_QUERY_POOL = "IMPALA_QUERY_POOL"


class ImpalaPluginConfiguration:
    def __init__(self, config):
        self.__config = config

    def host(self):
        return self.__config.get_value(IMPALA_HOST)

    def port(self):
        return self.__config.get_value(IMPALA_PORT)

    def database(self):
        return self.__config.get_value(IMPALA_DATABASE)

    def timeout(self):
        return self.__config.get_value(IMPALA_TIMEOUT)

    def use_ssl(self):
        return self.__config.get_value(IMPALA_USE_SSL)

    def ca_cert(self):
        return self.__config.get_value(IMPALA_CA_CERT)

    def auth_mechanism(self):
        return self.__config.get_value(IMPALA_AUTH_MECHANISM)

    def user(self):
        return self.__config.get_value(IMPALA_USER)

    def password(self):
        return self.__config.get_value(IMPALA_PW)

    def kerberos_service_name(self):
        return self.__config.get_value(IMPALA_KERBEROS_SERVICE_NAME)

    def krb_host(self):
        return self.__config.get_value(IMPALA_KRB_HOST)

    def use_http_transport(self):
        return self.__config.get_value(IMPALA_USE_HTTP_TRANSPORT)

    def http_path(self):
        return self.__config.get_value(IMPALA_HTTP_PATH)

    def auth_cookie_names(self):
        return self.__config.get_value(IMPALA_AUTH_COOKIE_NAMES)

    def retries(self):
        return self.__config.get_value(IMPALA_RETRIES)

    def sync_ddl(self):
        return self.__config.get_value(IMPALA_SYNC_DDL)

    def query_pool(self):
        return self.__config.get_value(IMPALA_QUERY_POOL)


def add_definitions(config_builder: ConfigurationBuilder) -> None:
    """
    Here we define what configuration settings are needed for Impala with reasonable defaults
    """
    config_builder.add(
        key=IMPALA_HOST,
        default_value="localhost",
        description="The Impala host we want to connect to.",
    )
    config_builder.add(
        key=IMPALA_PORT,
        default_value=21050,
        description="The Impala port we want to connect to.",
    )
    config_builder.add(
        key=IMPALA_DATABASE,
        default_value=None,
        description="The Impala database that we will be running queries against.",
    )
    config_builder.add(
        key=IMPALA_TIMEOUT,
        default_value=None,
        description="Impala connection timeout in seconds.",
    )
    config_builder.add(
        key=IMPALA_USE_SSL,
        default_value=False,
        description="Enables SSL for your connection to the database.",
    )
    config_builder.add(
        key=IMPALA_CA_CERT,
        default_value=None,
        description=(
            "Local path to the the third-party CA certificate."
            "If SSL is enabled but the certificate is not specified, the server certificate will not be validated."
        ),
    )
    config_builder.add(
        key=IMPALA_AUTH_MECHANISM,
        default_value="NOSASL",
        description=(
            "The Impala authentication mechanism. `'NOSASL'` for unsecured Impala. "
            "`'PLAIN'` for unsecured Hive (because Hive requires the SASL transport). "
            "`'GSSAPI'` for Kerberos and `'LDAP'` for Kerberos with LDAP."
        ),
    )
    config_builder.add(
        key=IMPALA_USER, default_value=None, description="Your Impala username."
    )
    config_builder.add(
        key=IMPALA_PW,
        default_value=None,
        description="Your Impala password.",
    )
    config_builder.add(
        key=IMPALA_KERBEROS_SERVICE_NAME,
        default_value=None,
        description="Used to authenticate to a particular `impalad` service principal. Uses `'impala'` by default.",
    )
    config_builder.add(
        key=IMPALA_KRB_HOST, default_value=None, description="The Kerberos host."
    )
    config_builder.add(
        key=IMPALA_USE_HTTP_TRANSPORT,
        default_value=False,
        description="Enables http transport; default is binary transport.",
    )
    config_builder.add(
        key=IMPALA_HTTP_PATH,
        default_value="",
        description="The path in the http URL. Used only when `impala_use_http_transport` is True.",
    )
    config_builder.add(
        key=IMPALA_AUTH_COOKIE_NAMES,
        default_value=None,
        description=(
            "The list of possible names for the cookie used for cookie-based authentication. "
            "If the list of names contains one cookie name only, a str value can be specified instead of a list. "
            "Used only when `impala_use_http_transport` is True. By default 'impala_auth_cookie_names' is set to the "
            "list of auth cookie names used by Impala and Hive. If 'impala_auth_cookie_names' is explicitly set to "
            "an empty value (None, [], or ''), VDK won't attempt to do cookie based authentication. "
            "Currently cookie-based authentication is only supported for GSSAPI over http."
        ),
    )
    config_builder.add(
        key=IMPALA_RETRIES,
        default_value=3,
        description="The number of retries VDK will attempt when connecting to or running a query against Impala.",
    )
    config_builder.add(
        key=IMPALA_SYNC_DDL,
        default_value=False,
        description=(
            "If enabled, DDL operations will return only when the changes have been propagated to all other "
            "Impala nodes in the cluster. For more information, see "
            "https://impala.apache.org/docs/build/html/topics/impala_sync_ddl.html"
        ),
    )
    config_builder.add(
        key=IMPALA_QUERY_POOL,
        default_value=None,
        description="The name of the Impala pool to execute queries in.",
    )



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
import logging

from impala.dbapi import connect as impala_connect
from vdk.internal.builtin_plugins.connection.managed_connection_base import (
    ManagedConnectionBase,
)

_log = logging.getLogger(__name__)


class ImpalaConnection(ManagedConnectionBase):
    def __init__(
        self,
        host="localhost",
        port=21050,
        database=None,
        timeout=None,
        use_ssl=False,
        ca_cert=None,
        auth_mechanism="NOSASL",
        user=None,
        password=None,
        kerberos_service_name="impala",
        krb_host=None,
        use_http_transport=False,
        http_path="",
        auth_cookie_names=None,
        retries=3,
    ):
        super().__init__(_log)

        self._host = host
        self._port = port
        self._database = database
        self._timeout = timeout
        self._use_ssl = use_ssl
        self._ca_cert = ca_cert
        self._auth_mechanism = auth_mechanism
        self._user = user
        self._password = password
        self._kerberos_service_name = kerberos_service_name
        self._krb_host = krb_host
        self._use_http_transport = use_http_transport
        self._http_path = http_path
        self._auth_cookie_names = auth_cookie_names
        self._retries = retries

    def _connect(self):
        conn = impala_connect(
            host=self._host,
            port=self._port,
            database=self._database,
            timeout=int(self._timeout) if self._timeout else None,
            use_ssl=self._use_ssl,
            ca_cert=self._ca_cert,
            auth_mechanism=self._auth_mechanism,
            user=self._user,
            password=self._password,
            kerberos_service_name=self._kerberos_service_name,
            krb_host=self._krb_host,
            use_http_transport=self._use_http_transport,
            http_path=self._http_path,
            auth_cookie_names=self._auth_cookie_names,
            retries=self._retries,
        )

        return conn

    def execute_query(self, query: str):
        return super().execute_query(query)



# Copyright 2021 VMware, Inc.
# SPDX-License-Identifier: Apache-2.0
from logging import getLogger
from typing import cast
from typing import Type

from pydantic import BaseModel
from pydantic import ValidationError
from vdk.api.job_input import IJobInput
from vdk.internal.builtin_plugins.run.job_input import JobInput
from vdk.internal.core import errors
from vdk.plugin.impala.impala_helper import ImpalaHelper

log = getLogger(__name__)


class TemplateArgumentsValidator:
    TemplateParams: Type[BaseModel]

    def __init__(self) -> None:
        pass

    def get_validated_args(self, job_input: IJobInput, args: dict) -> dict:
        args.update(self._validate_args(args))
        args["_vdk_template_insert_partition_clause"] = ""

        impala_helper = ImpalaHelper(cast(JobInput, job_input).get_managed_connection())
        table_name = "`{target_schema}`.`{target_table}`".format(**args)
        table_description = impala_helper.get_table_description(table_name)
        partitions = impala_helper.get_table_partitions(table_description)
        if partitions:
            args[
                "_vdk_template_insert_partition_clause"
            ] = impala_helper.get_insert_sql_partition_clause(partitions)

        impala_helper.ensure_table_format_is_parquet(table_name, table_description)

        source_view_full_name = "`{source_schema}`.`{source_view}`".format(**args)
        raw_source_view_has_results = job_input.execute_query(
            """
            WITH limited_view AS (SELECT * FROM {} LIMIT 1)
            SELECT COUNT(1) > 0 FROM limited_view
            """.format(
                source_view_full_name
            )
        )
        source_view_has_results = raw_source_view_has_results[0][0]
        if not source_view_has_results:
            log.info("Source view returns no results. Will NOT execute template!")
            raise Exception(
                "Source view returns no results. Will NOT execute template!"
            )
        return args

    def _validate_args(self, args: dict) -> dict:
        try:
            return self.TemplateParams(**args).dict()
        except ValidationError as error:
            errors.log_and_rethrow(
                to_be_fixed_by=errors.ResolvableBy.USER_ERROR,
                log=log,
                what_happened="Template execution in Data Job finished with error",
                why_it_happened=errors.MSG_WHY_FROM_EXCEPTION(error),
                consequences=errors.MSG_CONSEQUENCE_TERMINATING_APP,
                countermeasures=errors.MSG_COUNTERMEASURE_FIX_PARENT_EXCEPTION,
                exception=error,
            )
